-- Turns the existing Event.Unit.Change.Remove event into something a little more addon-developer friendly.
-- The Event.Unit.Change events are designed for power. Using them, you can, quite literally, track the entire potentially-infinite tree of .target chaining.
-- Most of the time, you don't want to. This library provides a simpler abstraction for events.

-- Usage: Create a handler with Library.UnitChange("YourAddonIdentifier", {"a", "list", "of", "unit", "specs", "that", "you", "care", "about", "player", "player.target", "mouseover"}, function (changes) print("A function that gets called when those change, in key/value form where the key is the unit spec and the value is the unit ID.") end).
-- If you think you may want to turn that handler off, save the return value - it contains a function (under .Delete) which can be called, with zero parameters, to stop processing.
-- There is currently no way to add or remove things from an existing UnitChange implementation. Feel free to delete it and create a new one. As long as you remove the old ones, new ones won't stack up and cause performance issues.

-- The documentation is a bit sparse.

if not Library then Library = {} end

local function process(data, changes)
  -- figure out what's different
  local refreshes = {}
  for change in pairs(changes) do
    if data.id[change] then
      for element in pairs(data.id[change]) do
        refreshes[element] = true
        
        -- Add chains as well
        if data.spec[element] then
          for chain in pairs(data.spec[element]) do
            refreshes[chain] = true
          end
        end
      end
      
      data.id[change] = nil
    end
  end

  -- From here, we need to take every change and link it up to the new things it points to
  local newresults = Inspect.Unit.Lookup(refreshes)
  
  local outputs = {}
  
  for unitspec in pairs(refreshes) do
    local unitid = newresults[unitspec] or false
    if not data.id[unitid] then
      data.id[unitid] = {}
    end
    data.id[unitid][unitspec] = true
    
    if data.current[unitspec] ~= unitid then
      outputs[unitspec] = unitid
      data.current[unitspec] = unitid
    end
  end
  
  if next(outputs) then
    data.func(outputs)
  end
end

local function eventHandler(tables, changes)
  for k, v in pairs(tables) do
    process(k, changes)
  end
end

local registered = {}
function Library.UnitChange(identifier, unitlist, func)
  if not registered[identifier] then
    registered[identifier] = {}
    
    table.insert(Event.Unit.Change.Remove, {function (changes) eventHandler(registered[identifier], changes) end, identifier, "LibUnitChange"})
  end
  
  local newTable = {func = func, spec = {}, id = {[false] = {}}, current = {}}
  
  for _, unit in ipairs(unitlist) do
    newTable.id[false][unit] = true
    
    local acum = nil
    for subset in unit:gmatch("[^.]+") do
      if acum then
        acum = acum .. "." .. subset
      else
        acum = subset
      end
      
      if not newTable.spec[acum] then
        newTable.spec[acum] = {}
      end
      newTable.spec[acum][unit] = true
    end
    
    newTable.current[unit] = false
  end
  
  registered[identifier][newTable] = true
  
  process(newTable, {[false] = false}) -- It's a fake message, but it's one that will get us to poll *everything*, which is exactly what we need.
  
  return {Delete = function ()
    registered[identifier][newTable] = nil -- deallocate
  end}
end
